//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>
#include <cstdlib>
#include <cstring>
#include <libexplain/fclose.h>
#include <libexplain/fflush.h>
#include <libexplain/fopen.h>
#include <libexplain/fread.h>
#include <libexplain/fwrite.h>
#include <libexplain/program_name.h>
#include <libexplain/stat.h>
#include <libexplain/unlink.h>
#include <map>
#include <string>
#include <sys/stat.h>

#include <lib/string_from_format.h>

#include <itractorgen/cgi.h>
#include <itractorgen/grammar.h>
#include <itractorgen/style.h>


typedef std::map<std::string, std::string> names_t;

static names_t names;


static void
name_eq_value(const std::string &text)
{
    if (text.empty())
        return;
    const char *s = text.c_str();
    const char *eq = strchr(s, '=');
    if (eq)
    {
        std::string name(s, eq - s);
        ++eq;
        const char *end = s + text.size();
        std::string value(eq, end - eq);
        names[name] = value;
    }
    else
        names[text] = "1";
}


static void
parse_query_string(void)
{
    const char *cp = getenv("QUERY_STRING");
    if (!cp || !*cp)
        return;
    for (;;)
    {
        const char *ampersand = strchr(cp, '&');
        if (!ampersand)
        {
            name_eq_value(cp);
            break;
        }
        name_eq_value(std::string(cp, ampersand - cp));
        cp = ampersand + 1;
    }
}


static void
parse_path_info(void)
{
    const char *cp = getenv("PATH_INFO");
    if (!cp || !*cp)
        return;
    for (;;)
    {
        const char *slash = strchr(cp, '/');
        if (!slash)
        {
            name_eq_value(cp);
            break;
        }
        name_eq_value(std::string(cp, slash - cp));
        cp = slash + 1;
    }
}


static long
size_of_file(const std::string &filename)
{
    struct stat st;
    explain_stat_or_die(filename.c_str(), &st);
    return st.st_size;
}


void
cgi(void)
{
    names["style"] = "random";
    names["factor"] = "random";
    names["maximum"] = "64";
    parse_query_string();
    parse_path_info();

    //
    // Get the style to be used.  Cope with "random".
    //
    std::string style = names["style"];
    if (style == "random")
        style = random_style_name();
    std::string filename = filename_from_style_name(style);

    std::string outfilename =
        string_from_format
        (
            "/tmp/%s-%ld.png",
            explain_program_name_get(),
            random()
        );

    //
    // Get the maximum height.
    //
    int maximum_height = 64;
    {
        std::string value = names["maximum"];
        maximum_height = strtol(value.c_str(), 0, 10);
        if (maximum_height < 10)
            maximum_height = 10;
        else if (maximum_height > 512)
            maximum_height = 512;
    }

    //
    // Get the tractor factor to be used.  Cope with "random".
    //
    int factor = 4;
    {
        std::string value = names["factor"];
        if (value == "random")
        {
            // factor will be in the range 4..9
            factor = random() % 6 + 4;
        }
        else
        {
            factor = strtol(value.c_str(), 0, 10);
            if (factor < 4)
                factor = 4;
            else
            {
                // We don't want all the CPU consumed by some genius
                // figuring out how to get a factor=1000000 image.
                if (factor > 9)
                    factor = 9;
            }
        }
    }

    //
    // Turn the tractor definition file into an image of a tractor.
    //
    file_from_style(filename, factor, maximum_height, outfilename);

    //
    // Now we have to emit suitably formatted output, mostly this
    // consists of writing an acceptable CGI header, followed by the
    // binary data of the output image.
    //
    printf("Status: 200 OK\n");
    printf("Content-Type: image/png\n");
    printf("Content-Length: %ld\n", size_of_file(outfilename));
    printf("\n");
    FILE *fp = explain_fopen_or_die(outfilename.c_str(), "rb");
    for (;;)
    {
        char data[1 << 12];
        size_t n = explain_fread_or_die(data, 1, sizeof(data), fp);
        if (!n)
            break;
        explain_fwrite_or_die(data, 1, n, stdout);
    }
    explain_fclose_or_die(fp);
    explain_fflush_or_die(stdout);
    explain_unlink_or_die(outfilename.c_str());
}
