//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef ITRACTORGEN_EXPRESSION_H
#define ITRACTORGEN_EXPRESSION_H

#include <lib/config.h>
#include <boost/shared_ptr.hpp>

#include <lib/gcc_attributes.h>
#include <lib/image.h>
#include <lib/rectangle.h>

#include <itractorgen/position.h>

/**
  * The expression class is used to represent a portion of the abstract
  * syntax tree describing how to build a tractor.
  */
class expression
{
public:
    typedef boost::shared_ptr<expression> pointer;

    /**
      * The destructor.
      */
    virtual ~expression();

    /**
      * The function_factory class method may be used to create anew
      * dynamically allocated instance of an expression derived class,
      * based on its name.  This is used by the grammar to resolve
      * function calls.
      *
      * @param name
      *     The name of the function (usually closely related to the
      *     name of the class).
      * @param arguments
      *     The function araguments.  Will be checked before used.
      */
    static pointer function_factory(const std::string &name,
        const pointer &arguments);

    /**
      * The is_an_image method is used to determine whether or not the
      * expression is able to be evaluated as an image.
      */
    virtual bool is_an_image(void) const = 0;

    /**
      * The is_a_number method is used to determine whether or not the
      * expression is able to be evaluated as a numeric value.
      */
    virtual bool is_a_number(void) const = 0;

    /**
      * The get_bounding_box method may be used to obtain the bounding
      * box of the expression, assuming it is an image.  Not image
      * expressions return an empty rectangle.
      */
    virtual rectangle get_bounding_box(void) const = 0;

    /**
      * The get_body_region method may be used to obtain the subset
      * of the image that is signifcant in placement decisions when
      * assembling the complete tractor.  By default, this is the whole
      * bounding box.
      */
    virtual rectangle get_body_region(void) const = 0;

    /**
      * The draw method is used to draw the expression onto an output
      * image.  All necessary calculations are evaluated and resolved.
      *
      * @param img
      *     The image to be drawn upon.
      * @param anti_alias
      *     this will be zero for no anti-alias, and more than zero
      *     if an n-by-n grid of super samples should be averaged.
      */
    void draw(const image::pointer &img, unsigned anti_alias) const;

    /**
      * The get_pixel method may be used to evaluate the pixel at the
      * given coordinates.
      *
      * @param x
      *     the X coordinate of the position
      * @param y
      *     the Y coordinate of the position
      */
    virtual icon_pixel get_pixel(double x, double y) const = 0;

    /**
      * The super_sample method is used to calculate a pixel value by
      * super sampling within the pixel boundaries.
      *
      * @param x
      *     The X coordinate of the pixel.
      * @param y
      *     The Y coordinate of the pixel.
      * @param grid
      *     The gride size for the super sampling grid.  There will be
      *     grid**2 samples taken.
      */
    icon_pixel super_sample(int x, int y, unsigned grid) const;

    /**
      * The representation method may be used to obtain a textual
      * representation of the expression.  Mostlyof use when debugging.
      */
    virtual std::string representation(void) const = 0;

    /**
      * The fatal_error method is used to report a problem with this
      * expression, and fails immediately.
      *
      * @param fmt
      *     the format of the output, and it also defines the numer and
      *     types of the remaining arguments.
      */
    void fatal_error(const char *fmt, ...) const
                                                   NORETURN FORMAT_PRINTF(2, 3);

    /**
      * The lex_error method is used to report a problem with this
      * expression, but does not fail immediately.
      *
      * @param fmt
      *     the format of the output, and it also defines the numer and
      *     types of the remaining arguments.
      */
    void lex_error(const char *fmt, ...) const
                                                            FORMAT_PRINTF(2, 3);

    /**
      * The get_numeric_value method may be used to obtain the numeric
      * value of this expression, provided that it can be evaluated in
      * that way.
      */
    virtual double get_numeric_value(void) const = 0;

    /**
      * The resolve_alternatives method is used to resolve any random
      * elements in the expression tree into concrete choices.
      * This shall always be done before the expression's value is required.
      */
    virtual pointer resolve_alternatives(void) const = 0;

protected:
    /**
      * The default constructor.
      * For use by derived classes only.
      */
    expression();

private:
    /**
      * The pos instance variable is used to remember the position of
      * the expression within the source file.
      */
    position pos;

    /**
      * The copy constructor.  Do not use.
      */
    expression(const expression &);

    /**
      * The assignment operator.  Do not use.
      */
    expression &operator=(const expression &);
};

// vim: set ts=8 sw=4 et :
#endif // ITRACTORGEN_EXPRESSION_H
