//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef ITRACTORGEN_EXPRESSION_BODY_REGION_H
#define ITRACTORGEN_EXPRESSION_BODY_REGION_H

#include <itractorgen/expression.h>

class expression_comma; // forward

/**
  * The expression_body_region class is used to represent a function
  * that alters the body region of an image, rather than its default
  * body region being the same as the bounding box.
  */
class expression_body_region:
    public expression
{
public:
    /**
      * The destructor.
      */
    virtual ~expression_body_region();

    /** 
      * The create class method is used to create new dynamically
      * allocated instances of this class.
      *
      * @param deeper
      *     The image to be filtered.
      * @param region
      *     The new body region to be used.
      */ 
    static pointer create(const expression::pointer &deeper, 
        const rectangle &region);

    /** 
      * The create_func class method is used to create new dynamically
      * allocated instances of this class, based on functions arguments
      * seen in the grammar.
      *
      * @param arguments
      *     The arguments to the function.  Don't worry, they are all
      *     checked before any are used.
      */ 
    static pointer create_func(const expression_comma *arguments);

protected:
    // See base class for documentation.
    bool is_an_image(void) const;

    // See base class for documentation.
    bool is_a_number(void) const;

    // See base class for documentation.
    rectangle get_bounding_box(void) const;

    // See base class for documentation.
    rectangle get_body_region(void) const;

    // See base class for documentation.
    icon_pixel get_pixel(double x, double y) const;

    // See base class for documentation.
    std::string representation(void) const;

    // See base class for documentation.
    double get_numeric_value(void) const;

    // See base class for documentation.
    pointer resolve_alternatives(void) const;

private:
    /**
      * The constructor.
      * It is private on purpose, use the #create class method instead.
      *
      * @param deeper
      *     The image to be filtered.
      * @param region
      *     The new body region to be used.
      */
    expression_body_region(const expression::pointer &deeper,
        const rectangle &region);

    /**
      * The deeper instance variable is used to remember the deeper
      * image to be filtered.
      */
    expression::pointer deeper;

    /**
      * The region instance variable is used to remember the preferred
      * body region.
      */
    rectangle region;

    /**
      * The default constructor.  Do not use.
      */
    expression_body_region();

    /**
      * The copy constructor.  Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    expression_body_region(const expression_body_region &rhs);

    /**
      * The assignment operator.  Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    expression_body_region &operator=(const expression_body_region &rhs);
};

// vim: set ts=8 sw=4 et :
#endif // ITRACTORGEN_EXPRESSION_BODY_REGION_H
