//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>
#include <cstring>

#include <lib/sizeof.h>

#include <itractorgen/expression/alternatives.h>
#include <itractorgen/expression/comma.h>
#include <itractorgen/expression/empty.h>
#include <itractorgen/expression/body_region.h>
#include <itractorgen/expression/offset.h>
#include <itractorgen/expression/over.h>
#include <itractorgen/expression/scale.h>
#include <itractorgen/lex.h>


static expression::pointer
shift_over(const expression::pointer &older, const expression::pointer &newer)
{
    rectangle older_region = older->get_body_region();
    rectangle newer_region = newer->get_body_region();

    // we adjust X so that older on the left touches newer on the right
    double ox =
        older_region.x + older_region.dx - newer_region.x;

    // we adjyst Y so that newer max-y is thet same as older max-y
    double oy =
        (older_region.y + older_region.dy) - (newer_region.y + newer_region.dy);

    return
        expression_over::create
        (
            expression_offset::create(newer, ox, oy),
            older
        );
}


static expression::pointer
beside(const expression_comma *args)
{
    if (args->size() < 2)
    {
        ::lex_error
        (
            "body_region: requires 2 or more arguments, not %d",
            int(args->size())
        );
        return expression_empty::create();
    }
    expression::pointer ep = args->get(0);
    for (size_t j = 1; j < args->size(); ++j)
    {
        ep = shift_over(ep, args->get(j));
    }
    return ep;
}


expression::pointer
expression::function_factory(const std::string &name,
    const expression::pointer &arguments_generic)
{
    expression_comma::pointer arguments =
        boost::dynamic_pointer_cast<expression_comma>(arguments_generic);
    assert(arguments);

    struct table_t
    {
        const char *name;
        expression::pointer (*create)(const expression_comma *args);
    };

    static const table_t table[] =
    {
        { "alternatives", &expression_alternatives::create_func },
        { "beside", beside },
        { "body_region", &expression_body_region::create_func },
        // { "crop", &expression_crop::create_func },
        { "offset", &expression_offset::create_func },
        { "translate", &expression_offset::create_func },
        { "over", &expression_over::create_func },
        // { "rotate", &expression_rotate::create_func },
        { "scale", &expression_scale::create_func },
    };

    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (0 == strcasecmp(name.c_str(), tp->name))
        {
            return (*tp->create)(arguments.get());
        }
    }

    ::lex_error("function \"%s\" unknown", name.c_str());
    return expression_empty::create();
}


// vim: set ts=8 sw=4 et :
