//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/string_from_format.h>

#include <itractorgen/expression/comma.h>
#include <itractorgen/expression/empty.h>
#include <itractorgen/expression/offset.h>
#include <itractorgen/lex.h>


expression_offset::~expression_offset()
{
}


expression_offset::expression_offset(
    const expression::pointer &a_img,
    double a_dx,
    double a_dy
) :
    img(a_img),
    dx(a_dx),
    dy(a_dy)
{
    if (!img->is_an_image())
        img->fatal_error("expression being offset is not an image");
}


expression::pointer
expression_offset::create(const expression::pointer &a_img, double a_dx, 
    double a_dy)
{
    if (a_dx == 0 && a_dy == 0)
        return a_img;
    return pointer(new expression_offset(a_img, a_dx, a_dy));
}


expression::pointer
expression_offset::create_func(const expression_comma *args)
{
    if (args->size() != 3)
    {
        ::lex_error
        (
            "body_region: requires 3 arguments, not %d",
            int(args->size())
        );
        return expression_empty::create();
    }
    expression::pointer a1 = args->get(0);
    if (!a1->is_an_image())
    {
        a1->lex_error("body_region: argument 1: not an image");
        return expression_empty::create();
    }

    expression::pointer a2 = args->get(1);
    if (!a2->is_a_number())
    {
        a2->lex_error("body_region: argument 2: not a number");
        return expression_empty::create();
    }

    expression::pointer a3 = args->get(2);
    if (!a3->is_a_number())
    {
        a3->lex_error("body_region: argument 3: not a number");
        return expression_empty::create();
    }

    //
    // Now we can call the regular create class method.
    //
    return create(a1, a2->get_numeric_value(), a3->get_numeric_value());
}


bool
expression_offset::is_an_image(void)
    const
{
    return true;
}


bool
expression_offset::is_a_number(void)
    const
{
    return false;
}


rectangle
expression_offset::get_bounding_box(void)
    const
{
    rectangle r = img->get_bounding_box();
    r.x += dx;
    r.y += dy;
    return r;
}


rectangle
expression_offset::get_body_region(void)
    const
{
    rectangle r = img->get_body_region();
    r.x += dx;
    r.y += dy;
    return r;
}


icon_pixel
expression_offset::get_pixel(double x, double y)
    const
{
    return img->get_pixel(x - dx, y - dy);
}


std::string 
expression_offset::representation(void)
    const
{
    std::string result = "offset(";
    result += img->representation();
    result += string_from_format(", %g, %g)", dx, dy);
    return result;
}


double
expression_offset::get_numeric_value(void)
    const
{
    assert(!"should not be able to get to here");
    return 0;
}


expression::pointer
expression_offset::resolve_alternatives(void)
    const
{
    return create(img->resolve_alternatives(), dx, dy);
}


// vim: set ts=8 sw=4 et :
