//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef ITRACTORGEN_EXPRESSION_OFFSET_H
#define ITRACTORGEN_EXPRESSION_OFFSET_H

#include <itractorgen/expression.h>

class expression_comma; // forward

/**
  * The expression_offset class is used to represent an expression
  * that results in an image being offset (translated) to a differnt
  * position.
  */
class expression_offset:
    public expression
{
public:
    /**
      * The destructor.
      */
    virtual ~expression_offset();

    /** 
      * The create class method is used to create new dynamically
      * allocated instances of this class.
      *
      * @param img
      *     The image to be manipulated.
      * @param dx
      *     The amount to move the image along the X axis; may be negative.
      * @param dy
      *     The amount to move the image along the Y axis; may be negative.
      */ 
    static pointer create(const expression::pointer &img, double dx, double dy);

    /** 
      * The create_func class method is used to create new dynamically
      * allocated instances of this class, based on functions arguments
      * seen in the grammar.
      *
      * @param arguments
      *     The arguments to the function.  Don't worry, they are all
      *     checked before any are used.
      */ 
    static pointer create_func(const expression_comma *arguments);

protected:
    // See base class for documentation.
    bool is_an_image(void) const;

    // See base class for documentation.
    bool is_a_number(void) const;

    // See base class for documentation.
    rectangle get_bounding_box(void) const;

    // See base class for documentation.
    rectangle get_body_region(void) const;

    // See base class for documentation.
    icon_pixel get_pixel(double x, double y) const;

    // See base class for documentation.
    std::string representation(void) const;

    // See base class for documentation.
    double get_numeric_value(void) const;

    // See base class for documentation.
    pointer resolve_alternatives(void) const;

private:
    /**
      * The constructor.
      * It is private on purpose, use the #create class method instead.
      *
      * @param img
      *     The image to be manipulated.
      * @param dx
      *     The amount to move the image along the X axis; may be negative.
      * @param dy
      *     The amount to move the image along the Y axis; may be negative.
      */
    expression_offset(const expression::pointer &img, double dx, double dy);

    /**
      * The img instance variable is used to remember the image to be
      * manipulated.
      */
    expression::pointer img;

    /**
      * The dx instance variable is used to remember the amount to move
      * the image along the X axis; may be negative.
      */
    double dx;

    /**
      * The dy instance variable is used to remember the amount to move
      * the image along the Y axis; may be negative.
      */
    double dy;

    /**
      * The default constructor.  Do not use.
      */
    expression_offset();

    /**
      * The copy constructor.  Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    expression_offset(const expression_offset &rhs);

    /**
      * The assignment operator.  Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    expression_offset &operator=(const expression_offset &rhs);
};

// vim: set ts=8 sw=4 et :
#endif // ITRACTORGEN_EXPRESSION_OFFSET_H
