//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <itractorgen/expression/comma.h>
#include <itractorgen/expression/empty.h>
#include <itractorgen/expression/over.h>
#include <itractorgen/lex.h>


expression_over::~expression_over()
{
}


expression_over::expression_over(
    const expression::pointer &a_upper,
    const expression::pointer &a_lower
) :
    upper(a_upper),
    lower(a_lower)
{
    if (!upper->is_an_image())
        upper->fatal_error("left hand side of 'over' is not an image");
    if (!lower->is_an_image())
        lower->fatal_error("right hand side of 'over' is not an image");
}


expression_over::pointer
expression_over::create(const expression::pointer &a_upper, 
    const expression::pointer &a_lower)
{
    return pointer(new expression_over(a_upper, a_lower));
}


expression::pointer
expression_over::create_func(const expression_comma *args)
{
    if (args->size() != 2)
    {
        ::lex_error("over: requires 2 arguments, not %d", int(args->size()));
        return expression_empty::create();
    }

    expression::pointer a1 = args->get(0);
    if (!a1->is_an_image())
    {
        a1->lex_error("over: argument 1: not an image");
        return expression_empty::create();
    }

    expression::pointer a2 = args->get(1);
    if (!a2->is_an_image())
    {
        a2->lex_error("over: argument 2: not an image");
        return expression_empty::create();
    }

    //
    // Now we can use the ordinary create class method.
    //
    return create(a1, a2);
}


bool
expression_over::is_an_image(void)
    const
{
    return true;
}


bool
expression_over::is_a_number(void)
    const
{
    return false;
}


rectangle
expression_over::get_bounding_box(void)
    const
{
    rectangle r1 = upper->get_bounding_box();
    rectangle r2 = lower->get_bounding_box();
    return (r1 + r2);
}


rectangle
expression_over::get_body_region(void)
    const
{
    rectangle r1 = upper->get_body_region();
    rectangle r2 = lower->get_body_region();
    return (r1 + r2);
}


icon_pixel
expression_over::get_pixel(double x, double y)
    const
{
    icon_pixel a = upper->get_pixel(x, y);
    if (a.is_opaque())
        return a;
    icon_pixel b = lower->get_pixel(x, y);
    if (a.is_transparent())
        return b;
    if (b.is_transparent())
        return a;

    icon_pixel::real_type aa = a.get_alpha();
    icon_pixel::real_type ba = b.get_alpha();

    icon_pixel::real_type xa = aa + (1 - aa) * ba;
    icon_pixel::real_type inv_xa = 1. / xa;
    icon_pixel::real_type xr =
        (a.get_red() * aa + (1. - aa) * b.get_red() * ba) * inv_xa;
    icon_pixel::real_type xg =
        (a.get_green() * aa + (1. - aa) * b.get_green() * ba) * inv_xa;
    icon_pixel::real_type xb =
        (a.get_blue() * aa + (1. - aa) * b.get_blue() * ba) * inv_xa;

    icon_pixel result(xr, xg, xb, xa);
    return result;
}


std::string 
expression_over::representation(void)
    const
{
    return "(" + upper->representation() + " over " + lower->representation()
        + ")";
}


double
expression_over::get_numeric_value(void)
    const
{
    assert(!"should never reach here");
    return 0;
}


expression::pointer
expression_over::resolve_alternatives(void)
    const
{
    return create(upper->resolve_alternatives(), lower->resolve_alternatives());
}


// vim: set ts=8 sw=4 et :
