//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/string_from_format.h>

#include <itractorgen/expression/comma.h>
#include <itractorgen/expression/empty.h>
#include <itractorgen/expression/scale.h>


expression_scale::~expression_scale()
{
}


expression_scale::expression_scale(
    const pointer &a_deeper,
    double a_sx,
    double a_sy
) :
    deeper(a_deeper),
    sx(a_sx),
    sy(a_sy),
    isx(sx != 0 ? (1 / sx) : 0),
    isy(sy != 0 ? (1 / sy) : 0)
{
}


expression::pointer
expression_scale::create(const pointer &a_deeper, double a_sx, double a_sy)
{
    if (a_sx == 1 && a_sy == 1)
        return a_deeper;
    if (a_sx == 0 || a_sy == 0)
        return expression_empty::create();
    return pointer(new expression_scale(a_deeper, a_sx, a_sy));
}


expression::pointer
expression_scale::create_func(const expression_comma *args)
{
    if (args->size() != 2 && args->size() != 3)
    {
        args->lex_error
        (
            "function scale: requires 3 arguments (not %d)",
            int(args->size())
        );
        return expression_empty::create();
    }

    expression::pointer a1 = args->get(0);
    if (!a1->is_an_image())
    {
        a1->lex_error("function scale: argument 1: must be an image");
        return expression_empty::create();
    }

    expression::pointer a2 = args->get(1);
    if (!a2->is_a_number())
    {
        a2->lex_error("function scale: argument 2: must be a number");
        return expression_empty::create();
    }
    double sfx = a2->get_numeric_value();

    double sfy = sfx;
    if (args->size() >= 3)
    {
        expression::pointer a3 = args->get(1);
        if (!a3->is_a_number())
        {
            a3->lex_error("function scale: argument 3: must be a number");
            return expression_empty::create();
        }
        sfy = a3->get_numeric_value();
    }

    //
    // Now we can use the plain-vanilla create class method.
    //
    return create(a1, sfx, sfy);
}


rectangle
expression_scale::get_bounding_box(void)
    const
{
    rectangle r = deeper->get_bounding_box();
    r.x *= sx;
    r.y *= sy;
    r.dx *= sx;
    r.dy *= sy;
    return r;
}


rectangle
expression_scale::get_body_region(void)
    const
{
    rectangle r = deeper->get_body_region();
    r.x *= sx;
    r.y *= sy;
    r.dx *= sx;
    r.dy *= sy;
    return r;
}


icon_pixel
expression_scale::get_pixel(double x, double y)
    const
{
    return deeper->get_pixel(x * isx, y * isy);
}


bool
expression_scale::is_an_image(void)
    const
{
    return true;
}


bool
expression_scale::is_a_number(void)
    const
{
    return false;
}


std::string
expression_scale::representation(void)
    const
{
    std::string result = "scale(";
    result += deeper->representation();
    result += string_from_format(", %g, %g)", sx, sy);
    return result;
}


double
expression_scale::get_numeric_value(void)
    const
{
    assert(!"should not be able to reach here");
    return 0;
}


expression::pointer
expression_scale::resolve_alternatives(void)
    const
{
    return create(deeper->resolve_alternatives(), sx, sy);
}


// vim: set ts=8 sw=4 et :
