/*
 * itractorgen - generate images of tractors
 * Copyright (C) 2011 Peter Miller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

%{

#include <libexplain/output.h>
#include <map>

#include <itractorgen/grammar.h>
#include <itractorgen/lex.h>
#include <itractorgen/expression/comma.h>
#include <itractorgen/expression/empty.h>
#include <itractorgen/expression/image.h>
#include <itractorgen/expression/negate.h>
#include <itractorgen/expression/number.h>

#define YYDEBUG 1
#define YYERROR_VERBOSE 1

void
grammar_debug_enable(void)
{
    extern int yydebug;
    yydebug = 1;
}

typedef std::map<std::string, expression::pointer> variables_t;
static variables_t variables;


static void
define(const std::string &name, const expression::pointer &value)
{
    if (variables.find(name) != variables.end())
        lex_error("variable '%s' already defined", name.c_str());
    else
        variables.insert(variables_t::value_type(name, value));
}


void
grammar(const std::string &filename)
{
    extern int yyparse(void);
    lex_open(filename);
    variables.clear();
    define("empty", expression_empty::create());
    yyparse();
    lex_close();
}


expression::pointer
grammar_get(const std::string &name)
{
    expression::pointer ep = variables[name];
    if (!ep)
        explain_output_error_and_die("variable '%s' undefined", name.c_str());
    return ep;
}


expression::pointer
grammar_get_image(const std::string &name)
{
    expression::pointer ep = grammar_get(name);
    if (!ep->is_an_image())
    {
        explain_output_error_and_die
        (
            "variable '%s' is not an image",
            name.c_str()
        );
    }
    return ep;
}


%}

%token FILENAME
%token JUNK
%token NAME
%token NUMBER
%token STRING

%left NAME
%left '(' ')' '[' ']'
%right unary
%left white
%left '|'
%left '='
%left ','

%union
{
    std::string *lv_string;
    double lv_number;
    expression::pointer *lv_expr;
    expression_comma::pointer *lv_expr_list;
}

%type <lv_expr_list> arguments
%type <lv_expr> expression
%type <lv_string> FILENAME 
%type <lv_string> NAME 
%type <lv_number> NUMBER 
%type <lv_expr_list> optional_arguments
%type <lv_string> STRING 

%%

tractor
    : assignments
    ;

assignments
    : /* empty */
    | assignments assignment
    ;


assignment
    : NAME '=' expression ';'
        {
            std::string name(*$1);
            delete $1;
            expression::pointer value(*$3);
            delete $3;
            define(name, value);
        }
    | error
    ;

expression
    : NAME
        {
            std::string name(*$1);
            delete $1;
            expression::pointer ep = variables[name];
            if (!ep)
            {
                lex_error("variable '%s' undefined", name.c_str());
                ep = expression_empty::create();
            }
            $$ = new expression::pointer(ep);
        }
    | FILENAME
        {
            expression::pointer ep = expression_image::create(*$1);
            delete $1;
            $$ = new expression::pointer(ep);
        }
    | NUMBER
        {
            expression::pointer ep = expression_number::create($1);
            $$ = new expression::pointer(ep);
        }
/*
    | STRING
        {
            expression::pointer ep = expression_string::create(*$1);
            delete $1;
            $$ = new expression::pointer(ep);
        }
*/
    | NAME '(' optional_arguments ')'
        %prec NAME
        {
            expression::pointer ep = expression::function_factory(*$1, *$3);
            delete $1;
            delete $3;
            $$ = new expression::pointer(ep);
        }
    | '-' expression
        %prec unary
        {
            expression::pointer ep = expression_negate::create(*$2);
            delete $2;
            $$ = new expression::pointer(ep);
        }
/*
    | expression expression
        %prec white
        {
            expression::pointer ep = expression_sequence::create(*$1, *$2);
            delete $1;
            delete $2;
            $$ = new expression::pointer(ep);
        }
    | expression '|' expression
        {
            expression::pointer ep = expression_alternate::create(*$1, *$3);
            delete $1;
            delete $3;
            $$ = new expression::pointer(ep);
        }
*/
    ;

optional_arguments
    : /* empty */
        {
            expression_comma::pointer ep = expression_comma::create();
            $$ = new expression_comma::pointer(ep);
        }
    | arguments
        {
            $$ = $1;
        }
    ;

arguments
    : expression
        {
            expression_comma::pointer ep = expression_comma::create();
            ep->push_back(*$1);
            delete $1;
            $$ = new expression_comma::pointer(ep);
        }
    | arguments ',' expression
        {
            expression_comma::pointer ep = *$1;
            delete $1;
            ep->push_back(*$3);
            delete $3;
            $$ = new expression_comma::pointer(ep);
        }
    ;
