//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>
#include <cstdarg>
#include <cstring>
#include <libexplain/fclose.h>
#include <libexplain/fopen.h>
#include <libexplain/getc.h>
#include <libexplain/output.h>
#include <libexplain/ungetc.h>
#include <string>

#include <itractorgen/expression/comma.h>
#include <itractorgen/lex.h>
#include <itractorgen/grammar.yacc.h>


static std::string ifn;
static FILE *ifp;
static int line_number;
static int number_of_errors;


void
lex_open(const std::string &filename)
{
    ifn = filename;
    ifp = explain_fopen_or_die(ifn.c_str(), "r");
    line_number = 1;
    number_of_errors = 0;
}


void
lex_close(void)
{
    explain_fclose_or_die(ifp);
    ifp = 0;
    if (number_of_errors)
    {
        explain_output_error_and_die
        (
            "%s: found %d fatal error%s",
            ifn.c_str(),
            number_of_errors,
            (number_of_errors == 1 ? "" : "s")
        );
    }
}


int
grammar_lex(void)
{
    for (;;)
    {
        int c = explain_getc_or_die(ifp);
        if (c == EOF)
            return 0;
        switch (c)
        {
        case ' ':
        case '\f':
        case '\r':
        case '\t':
            break;

        case '\n':
            ++line_number;
            break;

        case '#':
            for (;;)
            {
                c = explain_getc_or_die(ifp);
                if (c == EOF)
                    break;
                if (c == '\n')
                {
                    ++line_number;
                    break;
                }
            }
            break;

        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
            {
                std::string result;
                for (;;)
                {
                    result += (char)c;
                    c = explain_getc_or_die(ifp);
                    if (c == EOF)
                        goto done;
                    switch ((unsigned char)c)
                    {
                    case '0': case '1': case '2': case '3': case '4':
                    case '5': case '6': case '7': case '8': case '9':
                        continue;

                    default:
                        break;
                    }
                    break;
                }
                if (c == '.')
                {
                    for (;;)
                    {
                        result += (char)c;
                        c = explain_getc_or_die(ifp);
                        if (c == EOF)
                            goto done;
                        switch ((unsigned char)c)
                        {
                        case '0': case '1': case '2': case '3': case '4':
                        case '5': case '6': case '7': case '8': case '9':
                            continue;

                        default:
                            break;
                        }
                        break;
                    }
                }
                explain_ungetc_or_die(c, ifp);
                done:
                grammar_lval.lv_number = strtod(result.c_str(), 0);
                return NUMBER;
            }

        case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G':
        case 'H': case 'I': case 'J': case 'K': case 'L': case 'M': case 'N':
        case 'O': case 'P': case 'Q': case 'R': case 'S': case 'T': case 'U':
        case 'V': case 'W': case 'X': case 'Y': case 'Z':
        case 'a': case 'b': case 'c': case 'd': case 'e': case 'f': case 'g':
        case 'h': case 'i': case 'j': case 'k': case 'l': case 'm': case 'n':
        case 'o': case 'p': case 'q': case 'r': case 's': case 't': case 'u':
        case 'v': case 'w': case 'x': case 'y': case 'z':
            {
                std::string result;
                bool is_a_file_name = false;
                for (;;)
                {
                    result += (char)c;
                    c = explain_getc_or_die(ifp);
                    if (c == EOF)
                        break;
                    switch ((unsigned char)c)
                    {
                    case '.': case '/': case '-':
                        is_a_file_name = true;
                        // fall through...

                    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
                    case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
                    case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
                    case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
                    case 'Y': case 'Z':
                    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
                    case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
                    case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
                    case 's': case 't': case 'u': case 'v': case 'w': case 'x':
                    case 'y': case 'z':
                    case '0': case '1': case '2': case '3': case '4':
                    case '5': case '6': case '7': case '8': case '9':
                    case '_':
                        continue;

                    default:
                        explain_ungetc_or_die(c, ifp);
                        break;
                    }
                    break;
                }
                grammar_lval.lv_string = new std::string(result);
                if (is_a_file_name)
                    return FILENAME;
                return NAME;
            }

        default:
            if (!isprint((unsigned char)c))
                return JUNK;
            return c;
        }
    }
}


void
grammar_error(const char *text)
{
    lex_error("%s", text);
}


void
lex_error(const char *fmt, ...)
{
    char buffer[2000];
    snprintf(buffer, sizeof(buffer), "%s: %d: ", ifn.c_str(), line_number);
    char *cp = buffer + strlen(buffer);
    char *end = buffer + sizeof(buffer);

    va_list ap;
    va_start(ap, fmt);
    vsnprintf(cp, end - cp, fmt, ap);
    va_end(ap);

    explain_output_message(buffer);
    ++number_of_errors;
}


position
lex_get_position(void)
{
    return position(ifn, line_number);
}


void
lex_one_more_error(void)
{
    ++number_of_errors;
}
