//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>
#include <getopt.h>
#include <libexplain/output.h>
#include <libexplain/program_name.h>
#include <unistd.h>

#include <lib/dirname.h>
#include <lib/libdir.h>
#include <lib/search_path.h>
#include <lib/version.h>

#include <itractorgen/cgi.h>
#include <itractorgen/grammar.h>
#include <itractorgen/movie.h>
#include <itractorgen/style.h>


static void
usage(void)
{
    const char *prog = explain_program_name_get();
    fprintf(stderr, "Usage: %s [ <option>... ]\n", prog);
    fprintf(stderr, "       %s --version\n", prog);
    fprintf(stderr, "\n");
    fprintf(stderr, "  --file <name>     - the input script file\n");
    fprintf(stderr, "  --output <name>   - the output image file\n");
    fprintf(stderr, "  --factor <number> - the tractor factor\n");
    fprintf(stderr, "  --version         - print this mostly unhelpful text\n");
    exit(1);
}


static void
too_many_input_files(void)
{
    explain_output_error_and_die("too many input files specified");
}


int
main(int argc, char **argv)
{
    explain_program_name_set(argv[0]);
    srandom(getpid() * 17 + time(0));
    search_path_init();
    if (getenv("QUERY_STRING"))
    {
        cgi();
        return 0;
    }
    std::string filename;
    const char *outfilename = 0;
    int factor = 4;
    int maximum_height = 64;
    for (;;)
    {
        static struct option options[] =
        {
            { "animate", 1, 0, 'A' },
            { "configured", 0, 0, 'C' },
            { "factor", 1, 0, 'n' },
            { "file", 1, 0, 'f' },
            { "height", 1, 0, 'M' },
            { "input-file", 1, 0, 'f' },
            { "list", 0, 0, 'l' },
            { "maximum-height", 1, 0, 'M' },
            { "movie", 1, 0, 'A' },
            { "output", 1, 0, 'f' },
            { "output-file", 1, 0, 'o' },
            { "search-path", 1, 0, 'P' },
            { "style", 1, 0, 's' },
            { "tractor-factor", 1, 0, 'n' },
            { "version", 0, 0, 'V' },
            { "yacc-debug", 0, 0, 'y' },
            { 0, 0, 0, 0 }
        };
        int c = getopt_long(argc, argv, "A:Cf:ln:M:o:P:s:Vy", options, 0);
        if (c < 0)
            break;
        switch (c)
        {
        case 'A':
            movie(optarg);
            return 0;

        case 'C':
            printf("bindir=%s\n", configured_bindir().c_str());
            printf("datadir=%s\n", configured_datadir().c_str());
            printf("libdir=%s\n", configured_libdir().c_str());
            return 0;

        case 'f':
            if (!filename.empty())
                too_many_input_files();
            filename = optarg;
            search_path.push_front(dirname(filename));
            break;

        case 'l':
            style_list();
            return 0;

        case 'M':
            {
                char *ep = 0;
                maximum_height = strtol(optarg, &ep, 10);
                if (ep == optarg || *ep)
                {
                    explain_output_error_and_die
                    (
                        "maximum height \"%s\" does not appear to be a number",
                        optarg
                    );
                }
                if (maximum_height < 8)
                {
                    // Rather than being consistent with kfish's original
                    // tractorgen(1), which exits via abort(2), we will just
                    // be childish.
                    explain_output_error_and_die
                    (
                        "maximum height of %d is out of range, so there",
                        maximum_height
                    );
                }
            }
            break;

        case 'n':
            {
                if (0 == strcasecmp(optarg, "random"))
                {
                    factor = random() % 4 + 5;
                    break;
                }
                char *ep = 0;
                factor = strtol(optarg, &ep, 10);
                if (ep == optarg || *ep)
                {
                    explain_output_error_and_die
                    (
                        "tractor factor \"%s\" does not appear to be a number",
                        optarg
                    );
                }
                if (factor < 4 || factor > 100)
                {
                    // The original tractorgen(1) by kfish would exit via
                    // abort(2) in these circumstances.  We will settle for
                    // just being irritating.
                    explain_output_error_and_die
                    (
                        "a tractor factor of %d is out of range, but "
                            "I refuse to tell you what the valid range "
                            "actually is",
                        factor
                    );
                }
            }
            break;

        case 'o':
            outfilename = optarg;
            break;

        case 'P':
            search_path.push_front(optarg);
            break;

        case 's':
            if (!filename.empty())
                too_many_input_files();
            if (0 == strcasecmp(optarg, "random"))
                filename = filename_from_random_style();
            else
                filename = filename_from_style_name(optarg);
            break;

        case 'V':
            print_version();
            return 0;

        case 'y':
            grammar_debug_enable();
            break;

        default:
            usage();
        }
    }
    if (filename.empty())
    {
        filename = filename_from_random_style();
    }
    if (!outfilename)
    {
        explain_output_error
        (
            "you must specify an output image file name, "
                "use the --output option"
        );
        usage();
    }
    if (optind < argc)
        usage();

    file_from_style(filename, factor, maximum_height, outfilename);
    return 0;
}


// vim: set ts=8 sw=4 et :
