//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <cstdio>
#include <cstdlib>
#include <libexplain/mkdir.h>
#include <libexplain/output.h>
#include <libexplain/program_name.h>
#include <libexplain/system.h>

#include <lib/ends_with.h>
#include <lib/image/format/png.h>
#include <lib/path_join.h>
#include <lib/progress.h>
#include <lib/sizeof.h>
#include <lib/string_from_format.h>

#include <itractorgen/expression/color.h>
#include <itractorgen/expression/crop.h>
#include <itractorgen/expression/offset.h>
#include <itractorgen/expression/over.h>
#include <itractorgen/movie.h>
#include <itractorgen/style.h>


static const unsigned fps = 24;
static const int super_sample = 3;


static std::string
hhmmss(unsigned long x)
{
    unsigned ff = x % fps;
    x /= fps;
    unsigned ss = x % 60;
    x /= 60;
    unsigned mm = x % 60;
    x /= 60;
    unsigned hh = x;

    return string_from_format("%02dh%02dm%02ds%02d", hh, mm, ss, ff);
}


static inline int
iround(double x)
{
    return floor(0.5 + x);
}


void
movie(const std::string &outfilename)
{
    if (!ends_with(outfilename, ".avi"))
    {
        explain_output_error_and_die
        (
            "%s: must end with a .avi file extension",
            outfilename.c_str()
        );
    }

    const unsigned max_row_ht = 64;
    const unsigned width = 640;
    const unsigned height = 480;
    const unsigned nrows = height / max_row_ht;
    const unsigned start_y = (height - nrows * max_row_ht) / 2;
    const unsigned long nframes = fps * 60 * 60;
    std::string work_dir =
        string_from_format("/tmp/%s-%ld", explain_program_name_get(), random());

    struct row_t
    {
        image::pointer tractor;
        double offset;
        double pixels_per_frame;
        double ht;
        rectangle bb;
        row_t(): offset(0), pixels_per_frame(1), ht(64) { }

        void
        start_again(void)
        {
            std::string style_filename = filename_from_random_style();
            int tractor_factor = random() % 6 + 4;
            expression::pointer ep =
                expr_from_style(style_filename, tractor_factor, ht);
            bb = ep->get_bounding_box();
            tractor = image_format_png::create_blank(bb.idx(), bb.idy());
            ep->draw(tractor, super_sample);
            pixels_per_frame = (random() % 1000) / 250. + 0.5;
            offset = -bb.dx;
        }
    };
    row_t row[nrows];
    for (row_t *rp = row; rp < ENDOF(row); ++rp)
    {
        // fake perspective
        const size_t nrowm1 = SIZEOF(row) - 1;
        size_t idx = rp - row;
        rp->ht = max_row_ht * (0.9 + idx / (double)nrowm1);

        rp->start_again();

        if (width - rp->offset > nframes * rp->pixels_per_frame)
            rp->pixels_per_frame = (width - rp->offset) / nframes;
    }

    explain_mkdir_or_die(work_dir.c_str(), 0700);

    static icon_pixel forest_green(34/255., 139/255.,  34/255., 1.);
    image::pointer bg = image_format_png::create_blank(width, height);
    bg->set_pen_color(forest_green);
    bg->draw_rectangle(0, 0, width, height);

    progress verbose(0, nframes);
    verbose.begin();
    for (unsigned long j = 0; j < nframes; ++j)
    {
        std::string path =
            path_join(work_dir, hhmmss(j) + ".png");
        verbose.middle(j);

        image::pointer ip = image::create_blank_from_file(path, width, height);
        ip->copy(bg);

        for (row_t *rp = row; rp < ENDOF(row); ++rp)
        {
            double x = rp->offset;
            double y = (1 + rp - row) * max_row_ht + start_y - rp->bb.dy;
            ip->overlay(rp->tractor, iround(x), iround(y));
            rp->offset += rp->pixels_per_frame;

            if (rp->offset >= width)
            {
                rp->start_again();

                // Every body's rear end crosses the finish line at the
                // same time at the end of the movie.
                if (width - rp->offset > (nframes - j) * rp->pixels_per_frame)
                    rp->pixels_per_frame = (width - rp->offset) / (nframes - j);
            }
        }

        ip->save_to_file(path);
    }
    verbose.end();

    std::string command = "mencoder";
    command += " mf://" + work_dir + "/*.png";
    command += " -mf fps=" + string_from_format("%d", fps);
    command += " -o " + outfilename;
    command += " -ovc lavc -lavcopts vcodec=mpeg4";
    explain_output_message(command.c_str());
    explain_system_success_or_die(command.c_str());

    std::string command2 = "rm -rf " + work_dir;
    explain_output_message(command2.c_str());
    explain_system_success_or_die(command2.c_str());
}


// vim: set ts=8 sw=4 et :
