//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <libexplain/output.h>

#include <lib/dirstack/functor/list.h>
#include <lib/search_path.h>

#include <itractorgen/expression/comma.h>
#include <itractorgen/expression/offset.h>
#include <itractorgen/expression/over.h>
#include <itractorgen/expression/scale.h>
#include <itractorgen/grammar.h>
#include <itractorgen/style.h>


void
style_list(void)
{
    dirstack_functor_list f;
    search_path.for_each(f);
    f.list();
}


std::string
random_style_name(void)
{
    dirstack_functor_list f;
    search_path.for_each(f);
    std::string result = f.random();
    if (result.empty())
    {
        explain_output_error_and_die("could not find any built-in styles");
    }
    return result;
}


std::string
filename_from_style_name(const std::string &style_name)
{
    return search_path.find(style_name + ".tractor");
}


std::string
filename_from_random_style(void)
{
    return filename_from_style_name(random_style_name());
}


expression::pointer
expr_from_style(const std::string &filename, int factor, int maximum_height)
{
    unsigned super_sample = 0;
    return expr_from_style(filename, factor, maximum_height, super_sample);
}


expression::pointer
expr_from_style(const std::string &filename, int factor, int maximum_height,
    unsigned &super_sample)
{
    //
    // Parse the tractor definition.
    //
    grammar(filename);

    //
    // Get the three pieces of the tractor.
    //
    expression::pointer lhs = grammar_get_image("left");
    expression::pointer middle = grammar_get_image("middle");
    expression::pointer rhs = grammar_get_image("right");

    //
    // Glue the pieces together from left to right.
    //
    expression_comma::pointer args = expression_comma::create();
    args->push_back(lhs->resolve_alternatives());
    for (int j = 4; j < factor; ++j)
        args->push_back(middle->resolve_alternatives());
    args->push_back(rhs->resolve_alternatives());
    expression::pointer final = expression::function_factory("beside", args);
    args.reset();

    //
    // Make sure there is actually something there.
    //
    rectangle r = final->get_bounding_box();
    if (r.x != 0 || r.y != 0)
    {
        final = expression_offset::create(final, -r.x, -r.y);
        r = final->get_bounding_box();
    }
    if (r.dx < 1 || r.dy < 1)
        final->fatal_error("output image too small");

    // auto scale
    super_sample = 0;
    if (r.dy > maximum_height)
    {
        double sf = maximum_height / r.dy;
        final = expression_scale::create(final, sf, sf);
        r = final->get_bounding_box();

        super_sample = ceil(2 / sf);
        if (super_sample > 5)
            super_sample = 5;
        else if (super_sample <= 1)
            super_sample = 0;
    }
    return final;
}


void
file_from_style(const std::string &filename, int factor, int maximum_height,
    const std::string &outfilename)
{
    unsigned super_sample = 0;
    expression::pointer final =
        expr_from_style(filename, factor, maximum_height, super_sample);
    rectangle r = final->get_bounding_box();

    //
    // Draw the expression onto the output file.
    //
    image::pointer ip =
        image::create_blank_from_file(outfilename, r.idx(), r.idy());
    final->draw(ip, super_sample);
    ip->save_to_file(outfilename);
}
