//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_DIRSTACK_H
#define LIB_DIRSTACK_H

#include <lib/config.h>
#include <cstdio>
#include <list>
#include <string>

class dirstack_functor; // forward

/**
  * The dirstack class is used to represent an ordered list of
  * directories to be searched when attempting to open a file.
  */
class dirstack
{
public:
    /**
      * The destructor.
      * It is not virtual, thou shalt not derived from this class.
      */
    ~dirstack();

    /**
      * The default constructor.
      */
    dirstack();

    /**
      * The fopen method is used to search down the search path, looking
      * for the named file.
      *
      * @param filename
      *     the name of the file of interest.
      * @param mode
      *     the mode to be passed to the ::fopen function.
      * @returns
      *     pointer to the open file; it does not return in the file
      *     cannot be found instead it prints an error and exits.
      */
    FILE *fopen(const std::string &filename, const char *mode);

    /**
      * The find method is used to locate a file on the search path.
      *
      * @param filename
      *     the name of the file of interest.
      * @returns
      *     the path of the file including the relevant directory part;
      *     it returns the empty string if the file can't be found.
      */
    std::string find(const std::string &filename);

    /**
      * The push_back method is used to add a directory to the end of
      * the search path.
      */
    void push_back(const std::string &directory);

    /**
      * The push_front method is used to add a directory to the
      * beginning of the search path.
      */
    void push_front(const std::string &directory);

    /**
      * The for_each method is used to walk each directory in the path
      * in turn.
      *
      * @param f
      *     The functor to be called for each directory.
      */
    void for_each(dirstack_functor &f) const;

private:
    typedef std::list<std::string> directories_t;

    /**
      * The directories instance variable is used to remember the
      * ordered list of directories tro be searched when opening a file.
      */
    directories_t directories;

    /**
      * The copy constructor.  Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    dirstack(const dirstack &rhs);

    /**
      * The assignment operator.  Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    dirstack &operator=(const dirstack &rhs);
};

// vim: set ts=8 sw=4 et :
#endif // LIB_DIRSTACK_H
