//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_ICON_PIXEL_H
#define LIB_ICON_PIXEL_H

#include <lib/config.h>
#include <cmath>
#include <string>

/**
  * The icon_pixel class is used to represent a pixel of an RGBA image.
  *
  * (It isn't called plain "pixel" because of the anencephalic design
  * [actually, a distinct *lack* of design] of libnetpbm.  Name space
  * pollution everywhere!  Sheesh.)
  */
class icon_pixel
{
public:
    /**
      * The destructor.
      * It is not virtual, shou shalt not derive from this class.
      */
    ~icon_pixel(){}

    static const icon_pixel transparent;

    typedef double real_type;
    typedef unsigned short short_type;
    typedef unsigned char byte_type;

    /**
      * The default constructor.
      * The default pixel is completely transparent.
      */
    icon_pixel() :
        red_real(0),
        red_short(0),
        green_real(0),
        green_short(0),
        blue_real(0),
        blue_short(0),
        alpha_real(0),
        alpha_short(0),
        real_valid(true),
        short_valid(true)
    {
    }

    /**
      * The constructor.
      *
      * @param rr
      *     The red channel value, range [0, 65535].
      * @param gg
      *     The green channel value, range [0, 65535].
      * @param bb
      *     The blue channel value, range [0, 65535].
      *
      * The default alpha channel is completely opaque.
      */
    icon_pixel(short_type rr, short_type gg, short_type bb) :
        red_real(0),
        red_short(rr),
        green_real(0),
        green_short(gg),
        blue_real(0),
        blue_short(bb),
        alpha_real(1.),
        alpha_short(65535),
        real_valid(false),
        short_valid(true)
    {
    }

    /**
      * The constructor.
      *
      * @param rr
      *     The red channel value, range [0, 65535].
      * @param gg
      *     The green channel value, range [0, 65535].
      * @param bb
      *     The blue channel value, range [0, 65535].
      * @param aa
      *     The alpha channel value, range [0, 65535].
      */
    icon_pixel(short_type rr, short_type gg, short_type bb, short_type aa) :
        red_real(0),
        red_short(aa ? rr : 0),
        green_real(0),
        green_short(aa ? gg : 0),
        blue_real(0),
        blue_short(aa ? bb : 0),
        alpha_real(0),
        alpha_short(aa),
        real_valid(false),
        short_valid(true)
    {
    }

    /**
      * The constructor.
      *
      * @param rr
      *     The red channel value, range [0.0, 1.0].
      * @param gg
      *     The green channel value, range [0.0, 1.0].
      * @param bb
      *     The blue channel value, range [0.0, 1.0].
      *
      * The default alpha channel is completely opaque.
      */
    icon_pixel(real_type rr, real_type gg, real_type bb) :
        red_real(rr),
        red_short(0),
        green_real(gg),
        green_short(0),
        blue_real(bb),
        blue_short(0),
        alpha_real(1.0),
        alpha_short(65535),
        real_valid(true),
        short_valid(false)
    {
    }

    /**
      * The constructor.
      *
      * @param rr
      *     The red channel value, range [0.0, 1.0].
      * @param gg
      *     The green channel value, range [0.0, 1.0].
      * @param bb
      *     The blue channel value, range [0.0, 1.0].
      * @param aa
      *     The alpha channel value, range [0.0, 1.0].
      */
    icon_pixel(real_type rr, real_type gg, real_type bb, real_type aa) :
        red_real(aa > 0 ? rr : 0),
        red_short(0),
        green_real(aa > 0 ? gg : 0),
        green_short(0),
        blue_real(aa > 0 ? bb : 0),
        blue_short(0),
        alpha_real(aa),
        alpha_short(0),
        real_valid(true),
        short_valid(false)
    {
    }

    /**
      * The copy constructor.
      */
    icon_pixel(const icon_pixel &arg) :
        red_real(arg.red_real),
        red_short(arg.red_short),
        green_real(arg.green_real),
        green_short(arg.green_short),
        blue_real(arg.blue_real),
        blue_short(arg.blue_short),
        alpha_real(arg.alpha_real),
        alpha_short(arg.alpha_short),
        real_valid(arg.real_valid),
        short_valid(arg.short_valid)
    {
    }

    /**
      * The assignment operator.
      */
    icon_pixel &
    operator=(const icon_pixel &arg)
    {
        if (this != &arg)
        {
            red_real = arg.red_real;
            red_short = arg.red_short;
            green_real = arg.green_real;
            green_short = arg.green_short;
            blue_real = arg.blue_real;
            blue_short = arg.blue_short;
            alpha_real = arg.alpha_real;
            alpha_short = arg.alpha_short;
            real_valid = arg.real_valid;
            short_valid = arg.short_valid;
        }
        return *this;
    }

    /**
      * The is_transparent method is used to determine whether or not a
      * pixel is transparent, without incuring any on-demand conversion
      * penalty.
      */
    bool
    is_transparent(void)
        const
    {
        return (short_valid ? (alpha_short == 0) : (alpha_real < 0.002));
    }

    /**
      * The is_opaque method is used to determine whether or not a
      * pixel is opaque, without incuring any on-demand conversion
      * penalty.
      */
    bool
    is_opaque(void)
        const
    {
        return (short_valid ? (alpha_short >= 65408) : (alpha_real >= 1.0));
    }

    /**
      * The invert method may be used to create a new ixcon_pixel from
      * this one, where the color channels are inverted.  The alpha
      * channel retains the same value.
      */
    icon_pixel
    invert(void)
        const
    {
        return
            (
                real_valid
            ?
                icon_pixel
                (
                    1. - red_real,
                    1. - green_real,
                    1. - blue_real,
                    alpha_real
                )
            :
                icon_pixel
                (
                    short_type(65535 - red_short),
                    short_type(65535 - green_short),
                    short_type(65535 - blue_short),
                    alpha_short
                )
            );
    }

    real_type
    get_red(void)
        const
    {
        if (!real_valid)
            convert_short_to_real();
        return red_real;
    }

    real_type
    get_green(void)
        const
    {
        if (!real_valid)
            convert_short_to_real();
        return green_real;
    }

    real_type
    get_blue(void)
        const
    {
        if (!real_valid)
            convert_short_to_real();
        return blue_real;
    }

    real_type
    get_alpha(void)
        const
    {
        if (!real_valid)
            convert_short_to_real();
        return alpha_real;
    }

    short_type
    get_red_short(void)
        const
    {
        if (!short_valid)
            convert_real_to_short();
        return red_short;
    }

    byte_type
    get_red_byte(void)
        const
    {
        if (!short_valid)
            convert_real_to_short();
        return (red_short >> 8);
    }

    short_type
    get_green_short(void)
        const
    {
        if (!short_valid)
            convert_real_to_short();
        return green_short;
    }

    byte_type
    get_green_byte(void)
        const
    {
        if (!short_valid)
            convert_real_to_short();
        return (green_short >> 8);
    }

    short_type
    get_blue_short(void)
        const
    {
        if (!short_valid)
            convert_real_to_short();
        return blue_short;
    }

    byte_type
    get_blue_byte(void)
        const
    {
        if (!short_valid)
            convert_real_to_short();
        return (blue_short >> 8);
    }

    short_type
    get_alpha_short(void)
        const
    {
        if (!short_valid)
            convert_real_to_short();
        return alpha_short;
    }

    byte_type
    get_alpha_byte(void)
        const
    {
        if (!short_valid)
            convert_real_to_short();
        return (alpha_short >> 8);
    }

    /**
      * The representation method is used to obtain a representation of
      * this icon_pixel, usually for printing within debug messages.
      */
    std::string representation(void) const;

    /**
      * The equality operator is used to compare two pixel values to see
      * if they represent the same color.
      *
      * @param rhs
      *     The right hand side of the comparison.
      */
    bool operator==(const icon_pixel &rhs) const;

    /**
      * The inequality operator is used to compare two pixel values to
      * see if they represent the different colors.
      *
      * @param rhs
      *     The right hand side of the comparison.
      */
    bool
    operator!=(const icon_pixel &rhs)
        const
    {
        return !operator==(rhs);
    }

    /**
      * The multiply in situ operator.
      */
    void operator*=(const icon_pixel &rhs);

    /**
      * The multiply in situ operator.
      */
    icon_pixel operator*(const icon_pixel &rhs) const;

    /**
      * The multiply in situ operator.
      */
    void operator*=(double);

    /**
      * The multiply in situ operator.
      */
    icon_pixel operator*(double) const;

    /**
      * The monochrome method may be used to calculate a monochrime
      * pixel value, based on this pixlel.
      */
    icon_pixel monochrome(void) const;

    /**
      * The get_intensity is used to obtain the value of the Y channel
      * (the monochome channel).
      */
    real_type get_intensity(void) const;

    /**
      * The short_from_byte class method is used to turn a byte value into a
      * short value.
      */
    static short_type
    short_from_byte(byte_type x)
    {
        return ((short_type(x) << 8) | short_type(x));
    }

private:
    /**
      * The red_real instance variable is used to remember the value of
      * the Red channel, as a floating point number in the range [0.0,
      * 1.0].
      */
    mutable real_type red_real;

    /**
      * The red_short instance variable is used to remember the value of
      * the Red channel, as an integer in the range [0, 65535].
      */
    mutable short_type red_short;

    /**
      * The green_real instance variable is used to remember the value
      * of the Green channel, as a floating point number in the range
      * [0.0, 1.0].
      */
    mutable real_type green_real;

    /**
      * The green_short instance variable is used to remember the value
      * of the Green channel, as an integer in the range [0, 65535].
      */
    mutable short_type green_short;

    /**
      * The blue_real instance variable is used to remember the value of
      * the Blue channel, as a floating point number in the range [0.0,
      * 1.0].
      */
    mutable real_type blue_real;

    /**
      * The blue_short instance variable is used to remember the value
      * of the Blue channel, as an integer in the range [0, 65535].
      */
    mutable short_type blue_short;

    /**
      * The alpha_real instance variable is used to remember the value
      * of the Red channel, as a floating point number in the range
      * [0.0, 1.0].
      */
    mutable real_type alpha_real;

    /**
      * The alpha_short instance variable is used to remember the value
      * of the Alpha channel, as an integer in the range [0, 65535].
      */
    mutable short_type alpha_short;

    /**
      * The real_valid instance variable is used to remember whether or
      * not the *_real instance variables are valid.
      */
    mutable bool real_valid;

    /**
      * The short_valid instance variable is used to remember whether or
      * not the *_short instance variables are valid.
      */
    mutable bool short_valid;

    /**
      * The convert_channel_real_to_short class method is used to
      * convert a double in [0.0, 1.0] to an integer in [0, 65535] The
      * conversion is designed to divide the input range into 65536
      * equally-sized regions.  This is achieved by multiplying by 65536
      * and then special-casing the result of an input value of 1.0 so
      * that it maps to 65535 instead of 65536.
      */
    static short_type
    convert_channel_real_to_short(real_type x)
    {
        // multiply by 65535 is slow, so break it up
        unsigned n = x * 65536;
        n -= (n >> 16);
        return n;
    }

    /**
      * The convert_real_to_short method is used to translate the
      * realRGBA values into shortRGBA values.  This is done on demand.
      */
    void
    convert_real_to_short(void)
        const
    {
        red_short = convert_channel_real_to_short(red_real);
        green_short = convert_channel_real_to_short(green_real);
        blue_short = convert_channel_real_to_short(blue_real);
        alpha_short = convert_channel_real_to_short(alpha_real);
        short_valid = true;
    }

    /**
      * The convert_channel_short_to_real class method is used to
      * convert * an integer in [0, 65535] to a double in [0.0, 1.0].
      * This is achieved by dividing by 65536 and then special-casing
      * the result of an input value of 65535 so that it maps to 1.0
      * instead of 0.9999847412109375.
      */
    static real_type
    convert_channel_short_to_real(short_type x)
    {
        if (x >= 65535)
            return 1;
        return ldexp(real_type(x), -16);
    }

    void
    convert_short_to_real(void)
        const
    {
        red_real = convert_channel_short_to_real(red_short);
        green_real = convert_channel_short_to_real(green_short);
        blue_real = convert_channel_short_to_real(blue_short);
        alpha_real = convert_channel_short_to_real(alpha_short);
        real_valid = true;
    }
};

// vim: set ts=8 sw=4 et :
#endif // LIB_ICON_PIXEL_H
