//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_IMAGE_H
#define LIB_IMAGE_H

#include <lib/config.h>
#include <boost/shared_ptr.hpp>
#include <lib/icon_pixel.h>

/**
  * The image class is used to represent a generic image.
  */
class image
{
public:
    /**
      * The pointer type is to be used to point at any image instance.
      */
    typedef boost::shared_ptr<image> pointer;

    /**
      * The destructor.
      */
    virtual ~image();

    /**
      * The create_from_file class method is used to read a file and
      * create a new image by reading a file.
      *
      * @param filename
      *     The name of the file to be read
      */
    static pointer create_from_file(const std::string &filename);

    /**
     * The known_file_extension class method may be used to determine
     * whether or not the given file extension is a known image type.
     *
     * @param ext
     *     The file extension, without the dot.
     * @returns
     *     true if known, false if not
     */
    static bool known_file_extension(const std::string &ext);

    /**
      * The create_blank_from_file class method is used to examine the
      * file's extension and create a new image capable of writing that
      * format.  The file is not written now, you will have to call the
      * save_to_file method for that.
      *
      * @param filename
      *     The name of the file to be written, eventually.
      * @param width
      *     The width of the image in pixels.
      * @param height
      *     The height of the image in pixels.
      */
    static pointer create_blank_from_file(const std::string &filename,
        int width, int height);

    /**
      * The type_name method is used to obtain the name of the type of
      * this image format.
      */
    virtual const char *type_name(void) const = 0;

    /**
      * The copy method is used to copy the contents of another image
      * into this image.
      *
      * The copy is range checked.  If the other image is too small,
      * only the portion of this image covered by the other image will
      * be over-written.  If the other image is too large, only the
      * portion of the other image which covers this image will be used.
      *
      * @param src
      *     The image to copy.
      */
    void copy(const image::pointer &src);

    /**
      * The copy method is used to copy the contents of another image
      * into this image.
      *
      * The copy is range checked.  If the other image is too small,
      * only the portion of this image covered by the other image will
      * be over-written.  If the other image is too large, only the
      * portion of the other image which covers this image will be used.
      *
      * @param src
      *     The image to copy.
      *     @param ox
      * The X coordinate to start copying TO
      *     @param oy
      * The Y coordinate to start copying TO
      */
    void copy(const image::pointer &src, int ox, int oy);

    /**
      * The overlay method is used to copy the contents of another image
      * into this image, taking the alpha channel of the other image
      * into account.
      *
      * The operation is range checked.  If the other image is too
      * small, only the portion of this image covered by the other image
      * will be over-written.  If the other image is too large, only the
      * portion of the other image which covers this image will be used.
      *
      * @param src
      *     The image to copy.
      *     @param ox
      * The X coordinate to start copying TO
      *     @param oy
      * The Y coordinate to start copying TO
      */
    void overlay(const image::pointer &src, int ox, int oy);

    /**
      * The get_pixel method may be used to obtain the value of the
      * pixel at the given location.
      *
      * @param x
      *     The X cordinate of the pixel
      * @param y
      *     The Y cordinate of the pixel
      * @param value
      *     The place where the pixel value is to be stored on return.
      */
    virtual void get_pixel(int x, int y, icon_pixel &value) const = 0;

    /**
      * The get_pixel method may be used to obtain the value of the
      * pixel at the given location.
      *
      * @param x
      *     The X cordinate of the pixel
      * @param y
      *     The Y cordinate of the pixel
      * @returns
      *     the pixel value
      */
    icon_pixel
    get_pixel(int x, int y)
        const
    {
        icon_pixel result;
        get_pixel(x, y, result);
        return result;
    }

    /**
      * The set_pixel method may be used to set the value of the pixel
      * at the given location.  Transparency is ignored, the pixel
      * becomes exactly as given; if you want transparentcy taken into
      * account, use overlay_pixel.  The pen_clr tracks the given value.
      *
      * @param x
      *     The X coordinate of the pixel to be set
      * @param y
      *     The Y coordinate of the pixel to be set
      * @param value
      *     The value of the pixel to be set.
      */
    virtual void set_pixel(int x, int y, const icon_pixel &value) = 0;

    /**
      * The set_pixel method may be used to set the value of the pixel
      * at the given location.  Transparency is ignored, the pixel
      * becomes exactly as given; if you want transparentcy taken into
      * account, use overlay_pixel.  The pen_clr tracks the given value.
      *
      * @param x
      *     The X coordinate of the pixel to be set
      * @param y
      *     The Y coordinate of the pixel to be set
      */
    void set_pixel(int x, int y);

    /**
      * The overlay_pixel method may be used to set the value of the
      * pixel at the given location.  Transparency is taken into account
      * if the value has alpha of less than 255.
      *
      * @param x
      *     The X coordinate of the pixel to be set
      * @param y
      *     The Y coordinate of the pixel to be set
      * @param value
      *     The value of the pixel to be set.
      */
    virtual void overlay_pixel(int x, int y, const icon_pixel &value);

    /**
      * The overlay_pixel method may be used to set the value of the
      * pixel at the given location.  The pen_clr is used.  Transparency
      * is taken into account if pen_clr has alpha of less than 255.
      *
      * @param x
      *     The X coordinate of the pixel to be set
      * @param y
      *     The Y coordinate of the pixel to be set
      */
    virtual void overlay_pixel(int x, int y);

    /**
      * The get_width method may be used to obtain the width of the
      * image, in pixels.
      */
    virtual unsigned get_width(void) const = 0;

    /**
      * The get_height method may be used to obtain the height of the
      * image, in pixels.
      */
    virtual unsigned get_height(void) const = 0;

    /**
      * The save_to_file method is used to store the image data into a
      * file on disk.
      *
      * @param filename
      *     The name of the file to be written.
      */
    virtual void save_to_file(const std::string &filename) const = 0;

    void
    get_pen_position(int &x, int &y)
    {
        x = pen_x;
        y = pen_y;
    }

    void
    set_pen_position(int x, int y)
    {
        pen_x = x;
        pen_y = y;
    }

    void
    set_pen_color(const icon_pixel &clr)
    {
        pen_clr = clr;
    }

    /**
      * The draw_rectangle method is used to draw a simple aligned
      * rectangle on the image.
      *
      * @param x
      *     The left edge of the rectangle, in pixels
      * @param y
      *     The top edge of the rectangle, in pixels
      * @param w
      *     The width of the rectangle, in pixels
      * @param h
      *     The height of the ractangle, in pixels
      */
    void draw_rectangle(int x, int y, int w, int h);

    /**
      * The draw_rectangle method is used to draw a simple aligned
      * rectangle on the image.
      *
      * @param x
      *     The left edge of the rectangle, in pixels
      * @param y
      *     The top edge of the rectangle, in pixels
      * @param w
      *     The width of the rectangle, in pixels
      * @param h
      *     The height of the ractangle, in pixels
      * @param clr
      *     The color to draw the rectangle
      */
    void draw_rectangle(int x, int y, int w, int h, const icon_pixel &clr);

    /**
      * The same method is used to determine if two images are the
      * same as each other.  This is a size comparison, and only a
      * pixel-by-pixel comparison if the sizes are identical.
      */
    bool same(const image::pointer &rhs) const;

protected:
    /**
      * The default constructor.  May only be used by derived classes.
      */
    image();

private:
    int pen_x;
    int pen_y;
    icon_pixel pen_clr;
    // font::pointer pen_font;

    /**
      * The copy constructor.  Do not use.
      */
    image(const image &);

    /**
      * The assignment operator.  Do not use.
      */
    image &operator=(const image &);
};

// vim: set ts=8 sw=4 et :
#endif // LIB_IMAGE_H
