//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>
#include <cstring>
#include <libexplain/output.h>

// #include <lib/image/format/icns.h>
#include <lib/image/format/png.h>
#include <lib/image/format/gif.h>
// #include <lib/image/format/jpeg.h>
#include <lib/image/format/netpbm.h>
// #include <lib/image/format/tiff.h>
#include <lib/sizeof.h>


struct table_t
{
    const char *name;
    image::pointer (*create)(const std::string &filename);
    image::pointer (*create_blank)(int width, int height);
};


static table_t table[] =
{
    { "gif", &image_format_gif::create, 0 },
    // { "icns", &image_format_icns::create, 0 },
    // { "jpeg", &image_format_jpeg::create, &image_format_jpeg::create_blank },
    // { "jpg", &image_format_jpeg::create, &image_format_jpeg::create_blank },
    { "pam", &image_format_netpbm::create, &image_format_netpbm::create_blank },
    { "pbm", &image_format_netpbm::create, 0 },
    { "pgm", &image_format_netpbm::create, 0 },
    { "png", &image_format_png::create, &image_format_png::create_blank },
    { "pnm", &image_format_netpbm::create, &image_format_netpbm::create_blank },
    { "ppm", &image_format_netpbm::create, &image_format_netpbm::create_blank },
    // { "tif", &image_format_tiff::create, 0 },
    // { "tiff", &image_format_tiff::create, 0 },
};


static std::string
extn(const std::string &ss)
{
    const char *s = ss.c_str();
    const char *p = strchr(s, '/');
    p = p ? p + 1 : s;
    const char *q = strchr(p, '.');
    return (q ? q + 1 : "");
}


image::pointer
image::create_from_file(const std::string &filename)
{
    std::string extension = extn(filename);
    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (0 == strcasecmp(tp->name, extension.c_str()))
            return (*tp->create)(filename);
    }

    explain_output_error_and_die
    (
        "%s: not a recognised image format",
        filename.c_str()
    );
    return image_format_png::create(filename);
}


image::pointer
image::create_blank_from_file(const std::string &filename, int width,
    int height)
{
    std::string extension = extn(filename);
    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (0 == strcasecmp(tp->name, extension.c_str()))
        {
            if (!tp->create_blank)
            {
                explain_output_error_and_die
                (
                    "writing to a \"%s\" file is not supported",
                    extension.c_str()
                );
            }
            return (*tp->create_blank)(width, height);
        }
    }

    explain_output_error_and_die
    (
        "file extension \"%s\" is not in a recognised image format",
        extension.c_str()
    );
    return image_format_png::create_blank(width, height);
}


bool
image::known_file_extension(const std::string &extension)
{
    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (0 == strcasecmp(tp->name, extension.c_str()))
            return true;
    }
    return false;
}


// vim: set ts=8 sw=4 et :
