//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>
#include <cstring>
#include <gif_lib.h>
#include <libexplain/fclose.h>
#include <libexplain/fopen.h>
#include <libexplain/fread.h>
#include <libexplain/output.h>

#include <lib/image/format/gif.h>
#include <lib/search_path.h>


image_format_gif::~image_format_gif()
{
}


image_format_gif::image_format_gif(int a_width, int a_height, data_t *a_data) :
    image_format(a_width, a_height, a_data)
{
}


static const char *
gif_strerror(int x)
{
    switch (x)
    {
    case E_GIF_ERR_OPEN_FAILED:
        return "Failed to open given file";

    case E_GIF_ERR_WRITE_FAILED:
        return "Failed to Write to given file";

    case E_GIF_ERR_HAS_SCRN_DSCR:
        return "Screen Descriptor already been set";

    case E_GIF_ERR_HAS_IMAG_DSCR:
        return "Image Descriptor is still active";

    case E_GIF_ERR_NO_COLOR_MAP:
        return "Neither Global Nor Local color map";

    case E_GIF_ERR_DATA_TOO_BIG:
        return "#Pixels bigger than Width * Height";

    case E_GIF_ERR_NOT_ENOUGH_MEM:
        return "Fail to allocate required memory";

    case E_GIF_ERR_DISK_IS_FULL:
        return "Write failed (disk full?)";

    case E_GIF_ERR_CLOSE_FAILED:
        return "Failed to close given file";

    case E_GIF_ERR_NOT_WRITEABLE:
        return "Given file was not opened for write";

    case D_GIF_ERR_OPEN_FAILED:
        return "Failed to open given file";

    case D_GIF_ERR_READ_FAILED:
        return "Failed to Read from given file";

    case D_GIF_ERR_NOT_GIF_FILE:
        return "Given file is NOT GIF file";

    case D_GIF_ERR_NO_SCRN_DSCR:
        return "No Screen Descriptor detected";

    case D_GIF_ERR_NO_IMAG_DSCR:
        return "No Image Descriptor detected";

    case D_GIF_ERR_NO_COLOR_MAP:
        return "Neither Global Nor Local color map";

    case D_GIF_ERR_WRONG_RECORD:
        return "Wrong record type detected";

    case D_GIF_ERR_DATA_TOO_BIG:
        return "#Pixels bigger than Width * Height";

    case D_GIF_ERR_NOT_ENOUGH_MEM:
        return "Fail to allocate required memory";

    case D_GIF_ERR_CLOSE_FAILED:
        return "Failed to close given file";

    case D_GIF_ERR_NOT_READABLE:
        return "Given file was not opened for read";

    case D_GIF_ERR_IMAGE_DEFECT:
        return "Image is defective, decoding aborted";

    case D_GIF_ERR_EOF_TOO_SOON:
        return "Image EOF detected, before image complete";
    }
    return "undefined";
}


static int
input_func(GifFileType *gfp, GifByteType *data, int nbytes)
{
    if (nbytes <= 0)
        return 0;
    FILE *fp = (FILE *)(gfp->UserData);
    return explain_fread_or_die(data, nbytes, 1, fp);
}


image::pointer
image_format_gif::create(const std::string &filename)
{
    //
    // Initially there is no transparent color.
    //
    GifByteType transparent = -1;

    //
    // Open the file.
    //
    FILE *fp = search_path.fopen(filename.c_str(), "rb");
    GifFileType *g = DGifOpen(fp, input_func);
    if (!g)
    {
        explain_output_error_and_die
        (
            "open %s: %s",
            filename.c_str(),
            gif_strerror(GifLastError())
        );
    }

    //
    // We allocate space for GIF data, one byte per pixel.
    //
    size_t width = g->SWidth;
    size_t height = g->SHeight;
    size_t gif_size = width * height;
    GifByteType *gif_data = new GifByteType [gif_size];
    memset(gif_data, g->SBackGroundColor, gif_size);

    //
    // Read the data from the file.
    //
    for (;;)
    {
        GifRecordType RecordType;
        if (DGifGetRecordType(g, &RecordType) == GIF_ERROR)
        {
            oops:
            explain_output_error_and_die
            (
                "read %s: %s",
                filename.c_str(),
                gif_strerror(GifLastError())
            );
        }
        switch (RecordType)
        {
        case IMAGE_DESC_RECORD_TYPE:
            {
                if (DGifGetImageDesc(g) == GIF_ERROR)
                    goto oops;
                // Image Position relative to Screen.
                int sub_y = g->Image.Top;
                int sub_x = g->Image.Left;
                int sub_width = g->Image.Width;
                int sub_height = g->Image.Height;
                if
                (
                    g->Image.Left + g->Image.Width > g->SWidth
                ||
                    g->Image.Top + g->Image.Height > g->SHeight
                )
                {
                    explain_output_error_and_die
                    (
                        "%s: sub-image not confined to image dimension",
                        filename.c_str()
                    );
                }
                if (g->Image.Interlace)
                {
                    // The way Interlaced image should
                    // be read - offsets and jumps...
                    unsigned InterlacedOffset[4] = { 0, 4, 2, 1 };
                    unsigned InterlacedJumps[4] = { 8, 8, 4, 2 };

                    // Need to perform 4 passes on the images
                    for (int i = 0; i < 4; i++)
                    {
                        for
                        (
                            int j = sub_y + InterlacedOffset[i];
                            j < sub_y + sub_height;
                            j += InterlacedJumps[i]
                        )
                        {
                            GifByteType *pos = gif_data + width * j + sub_x;
                            int err = DGifGetLine(g, pos, sub_width);
                            if (err == GIF_ERROR)
                                goto oops;
                        }
                    }
                }
                else
                {
                    for (int i = 0; i < sub_height; i++)
                    {
                        GifByteType *pos =
                            gif_data + width * (sub_y + i) + sub_x;
                        int err = DGifGetLine(g, pos, sub_width);
                        if (err == GIF_ERROR)
                            goto oops;
                    }
                }
            }
            continue;

        case EXTENSION_RECORD_TYPE:
            {
                //
                // Skip any extension blocks in the file
                // (except for the transparent color)
                //
                int ExtCode = 0;
                GifByteType *Extension = 0;
                int err = DGifGetExtension(g, &ExtCode, &Extension);
                if (err == GIF_ERROR)
                    goto oops;
                if (ExtCode == GRAPHICS_EXT_FUNC_CODE)
                {
                    assert(!!Extension);
                    if (Extension[1] & 1)
                    {
                        transparent = (GifByteType)Extension[4];
                    }
                }
                while (Extension != NULL)
                {
                    err = DGifGetExtensionNext(g, &Extension);
                    if (err == GIF_ERROR)
                        goto oops;
                }
            }
            continue;

        case UNDEFINED_RECORD_TYPE:
        case SCREEN_DESC_RECORD_TYPE:
            assert(0);

        case TERMINATE_RECORD_TYPE:
            break;
        }
        break;
    }

    //
    // Use the color map to produce RGBA data.
    //
    ColorMapObject *ColorMap =
        (g->Image.ColorMap ? g->Image.ColorMap : g->SColorMap);
    size_t rgba_rowspan = width * 4;
    size_t rgba_size = height * rgba_rowspan;
    data_t *rgba_data = new data_t [rgba_size];
    for (size_t y = 0; y < height; ++y)
    {
        data_t *rgba_row = rgba_data + y * rgba_rowspan;
        GifByteType *gif_row = gif_data + y * width;
        for (size_t x = 0; x < width; ++x)
        {
            GifByteType c = *gif_row++;
            if (c == transparent)
            {
                *rgba_row++ = 0;
                *rgba_row++ = 0;
                *rgba_row++ = 0;
                *rgba_row++ = 0;
            }
            else
            {
                const GifColorType &ColorMapEntry = ColorMap->Colors[c];
                *rgba_row++ = ColorMapEntry.Red;
                *rgba_row++ = ColorMapEntry.Green;
                *rgba_row++ = ColorMapEntry.Blue;
                *rgba_row++ = 255;
            }
        }
    }
    delete [] gif_data;
    gif_data = 0;
    DGifCloseFile(g);
    g = 0;
    explain_fclose_or_die(fp);

    //
    // now create anew instance
    //
    return pointer(new image_format_gif(width, height, rgba_data));
}


const char *
image_format_gif::type_name(void)
    const
{
    return "GIF";
}


// vim: set ts=8 sw=4 et :
