//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>

extern "C" {
// NetPBM has the worst possible include file names
//
// And those horribly named include files contain equally horrible
// symbol names.  There is pollution in every possible C namespace.
#include <pam.h>
}

#include <cstring>
#include <libexplain/fclose.h>
#include <libexplain/fopen.h>
#include <libexplain/output.h>

#include <lib/image/format/netpbm.h>
#include <lib/search_path.h>


image_format_netpbm::~image_format_netpbm()
{
}


image_format_netpbm::image_format_netpbm(int a_width, int a_height) :
    image_format(a_width, a_height)
{
}


image_format_netpbm::image_format_netpbm(
    int a_width,
    int a_height,
    data_t *a_data
) :
    image_format(a_width, a_height, a_data)
{
}


const char *
image_format_netpbm::type_name(void)
    const
{
    return "NetPBM";
}


image::pointer
image_format_netpbm::create(const std::string &filename)
{
    //
    // open file and test for it being a netpbm file
    //
    std::string fn = filename;
    FILE *fp = search_path.fopen(fn, "rb");

    struct pam control;
    pnm_readpaminit(fp, &control, sizeof(control));
    if (control.maxval == 0)
        control.maxval = 255;
    tuple *tuplerow = pnm_allocpamrow(&control);
    int width = control.width;
    int height = control.height;
    size_t rgba_row_stride = width * 4;
    size_t rgba_data_size = rgba_row_stride * height;
    data_t *rgba_data = new data_t [rgba_data_size];

    //
    // read the image from the file
    //
    for (int y = 0; y < height; ++y)
    {
        data_t *op = rgba_data + y * rgba_row_stride;
        pnm_readpamrow(&control, tuplerow);
        switch (PAM_FORMAT_TYPE(control.format))
        {
        case PBM_TYPE:
            {
                assert(control.depth == 1);
                for (int x = 0; x < width; ++x)
                {
                    data_t n = (tuplerow[x][0] == PAM_PBM_BLACK ? 0 : 255);
                    *op++ = n;
                    *op++ = n;
                    *op++ = n;
                    *op++ = 255;
                }
            }
            break;

        case PGM_TYPE:
            {
                assert(control.depth == 1 || control.depth == 2);
                for (int x = 0; x < width; ++x)
                {
                    data_t n = tuplerow[x][0] * 255 / control.maxval;
                    *op++ = n;
                    *op++ = n;
                    *op++ = n;
                    *op++ =
                        (
                            control.bytes_per_sample == 2
                        ?
                            tuplerow[x][1] * 255 / control.maxval
                        :
                            255
                        );
                }
            }
            break;

        case PPM_TYPE:
        case PAM_TYPE:
            {
                assert(control.depth == 3 || control.depth == 4);
                for (int x = 0; x < width; ++x)
                {
                    *op++ = tuplerow[x][0] * 255 / control.maxval;
                    *op++ = tuplerow[x][1] * 255 / control.maxval;
                    *op++ = tuplerow[x][2] * 255 / control.maxval;
                    *op++ =
                        (
                            control.bytes_per_sample == 4
                        ?
                            tuplerow[x][3] * 255 / control.maxval
                        :
                            255
                        );
                }
            }
            break;

        default:
            explain_output_error_and_die
            (
                "%s: don't grok format type 0x%04X",
                fn.c_str(),
                PAM_FORMAT_TYPE(control.format)
            );
        }
    }
    pnm_freepamrow(tuplerow);
    explain_fclose_or_die(fp);

    return pointer(new image_format_netpbm(width, height, rgba_data));
}


void
image_format_netpbm::save_to_file(const std::string &filename)
    const
{
    //
    // create file
    //
    FILE *fp = explain_fopen_or_die(filename.c_str(), "wb");

    //
    // initialize stuff
    //
    struct pam control;
    control.maxval = 255;
    control.format = RPPM_FORMAT;
    control.width = get_width();
    control.height = get_height();
    control.depth = 3; // would 4 preserve our alpha channel?
    control.size = sizeof(control);
    control.len = sizeof(control);
    control.file = fp;
    control.bytes_per_sample = 1;
    pnm_writepaminit(&control);

    //
    // write each row to the output
    //
    tuple *tuplerow = pnm_allocpamrow(&control);
    for (int y = 0; y < control.height; ++y)
    {
        const data_t *ip = get_data(y);
        for (int x = 0; x < control.width; ++x)
        {
            tuplerow[x][0] = *ip++;
            tuplerow[x][1] = *ip++;
            tuplerow[x][2] = *ip++;
            if (control.depth == 4)
                tuplerow[x][3] = *ip;
            ++ip;
        }
        pnm_writepamrow(&control, tuplerow);
    }
    pnm_freepamrow(tuplerow);

    explain_fclose_or_die(fp);
}


image::pointer
image_format_netpbm::create_blank(int width, int height)
{
    return pointer(new image_format_netpbm(width, height));
}


// vim: set ts=8 sw=4 et :
