//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_IMAGE_FORMAT_PNG_H
#define LIB_IMAGE_FORMAT_PNG_H

#include <lib/image/format.h>

/**
  * The image_format_png class is used to represent an image in PNG
  * format.
  */
class image_format_png:
    public image_format
{
public:
    /**
      * The destructor.
      */
    virtual ~image_format_png();

    /**
      * The create class method is used to create an image by reading a
      * file.
      *
      * @param filename
      *     The name of the file to be read.
      */
    static pointer create(const std::string &filename);

    /**
      * The create_blank class method is used to create a new image
      * which is entirely blank (and transparent).
      *
      * @param width
      *     The width of the image
      * @param height
      *     The height of the image
      */
    static pointer create_blank(int width, int height);

protected:
    // See base class for documentation.
    const char *type_name(void) const;

    // See base class for documentation.
    void save_to_file(const std::string &filename) const;

private:
    /**
      * The calc_needs_alpha method is used by the #save_to_file method
      * to determine whether or not the alpha channel needs to be stored
      * into the output file (i.e. if any pixel has an alpha != 255).
      *
      * @returns
      *     true of the image needs an explict alpha channel, false if
      *     an explict (all alpha == 255) channel is sufficient.
      */
    bool calc_needs_alpha(void) const;

    /**
      * The calc_needs_rgb method is used by the #save_to_file method to
      * determine whether or not the data is grayscale (r == b == b for
      * all pixels).  Completely transparent pixels are ignored.
      *
      * @returns
      *     true if at least one pixel (r != g || r != b), false if all
      *     the pixels can be represented with a grayscale file.
      */
    bool calc_needs_rgb(void) const;

    /**
      * The default constructor.  Do not use.
      */
    image_format_png();

    /**
      * The constructor.
      *
      * @param width
      *     The width of the new image
      * @param height
      *     The height of the new image
      * @param data
      *     The data of the new image, indexed row then column then
      *     component (RGBA).
      */
    image_format_png(int width, int height, data_t *data);

    /**
      * The constructor.
      *
      * @param width
      *     The width of the new image
      * @param height
      *     The height of the new image
      */
    image_format_png(int width, int height);

    /**
      * The copy constructor.  Do not use.
      */
    image_format_png(const image_format_png &);

    /**
      * The assignment operator.  Do not use.
      */
    image_format_png &operator=(const image_format_png &);
};

// vim: set ts=8 sw=4 et :
#endif // LIB_IMAGE_FORMAT_PNG_H
