//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/time.h>
#include <time.h>
#include <termios.h>
#include <unistd.h>

#include <lib/progress.h>


progress::~progress()
{
    end();
}


static bool
stdout_is_tty(void)
{
    return !!isatty(1);
}


static int
get_width(void)
{
    const char *cp = getenv("COLS");
    if (cp)
    {
        int n = atoi(cp);
        if (n > 0)
            return n;
    }

    //
    // Use stdout attributes, if available.
    //
#ifdef TIOCGWINSZ
    struct winsize ws;
    if (ioctl(1, TIOCGWINSZ, &ws) == 0)
    {
        if (ws.ws_col > 0)
            return ws.ws_col;
    }
#endif
    return 80;
}


static int
calc_digits(int x)
{
    int result = 1;
    if (x < 0)
    {
        ++result;
        x = -x;
    }
    for (;;)
    {
        if (x < 10)
            return result;
        ++result;
        x /= 10;
    }
}


static int
calc_digits_range(int x1, int x2)
{
    int d1 = calc_digits(x1);
    int d2 = calc_digits(x2);
    return (d1 > d2 ? d1 : d2);
}


progress::progress(int a_min, int a_max) :
    show(stdout_is_tty()),
    width(get_width()),
    need_to_clear(false),
    minimum(a_min),
    maximum(a_max),
    digits(calc_digits_range(a_min, a_max)),
    start_time(0),
    next(0)
{
    //assert(minimum < maximum);
    if (minimum >= maximum)
        maximum = minimum + 1;
}


static double
get_time(void)
{
    struct timeval now;
    gettimeofday(&now, 0);
    return (now.tv_sec + 1e-6 * now.tv_usec);
}


void
progress::begin(void)
{
    start_time = get_time();
    need_to_clear = true;
}


void
progress::middle(int position)
{
    // assert(minimum <= position);
    // assert(position <= maximum);
    if (!show)
        return;
    if (!start_time)
        begin();
    if (position <= minimum)
        return;
    double now = get_time();
    if (now < next)
        return;
    next = now + 1;

    double percent = 100. * (position - minimum) / (maximum - minimum);
    printf("%7.2f%%, ", percent);

    if (minimum == 0)
    {
        printf("%*d of %d", digits, position, maximum);
    }
    else
    {
        printf("%d <= %*d <= %d", minimum, digits, position, maximum - 1);
    }

    double elapsed = now - start_time;
    if (elapsed > 10)
    {
        double per_second = (position - minimum) / elapsed;
        double total_seconds = (maximum - minimum) / per_second;
        double remaining_seconds = total_seconds - elapsed;
        if (remaining_seconds >= 1)
        {
            long sec = (long)remaining_seconds;
            long min = sec / 60;
            sec -= min * 60;
            long hr = min / 60;
            min -= hr * 60;
            printf(", ETA ");
            if (hr > 0)
                printf("%ldh%02ldm", hr, min);
            else if (min > 0)
                printf("%ldm%02lds", min, sec);
            else
                printf("%lds", sec);
        }
    }
    printf("\33[K\r");
    fflush(stdout);
    need_to_clear = true;
}


void
progress::end(void)
{
    clear();
}


void
progress::clear(void)
{
    if (need_to_clear)
    {
        need_to_clear = false;
        printf("\r\33[K");
        fflush(stdout);
        need_to_clear = false;
    }
}


void
progress::set_show(bool flag)
{
    show = flag;
}


// vim: set ts=8 sw=4 et :
