//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>
#include <cmath>

#include <lib/rectangle.h>
#include <lib/string_from_format.h>


rectangle::~rectangle()
{
}


rectangle::rectangle() :
    x(0),
    y(0),
    dx(0),
    dy(0)
{
}


rectangle::rectangle(double a_x, double a_y, double a_dx, double a_dy) :
    x(a_x),
    y(a_y),
    dx(a_dx),
    dy(a_dy)
{
}


rectangle::rectangle(const rectangle &arg) :
    x(arg.x),
    y(arg.y),
    dx(arg.dx),
    dy(arg.dy)
{
}


rectangle &
rectangle::operator=(const rectangle &arg)
{
    if (this != &arg)
    {
        x = arg.x;
        y = arg.y;
        dx = arg.dx;
        dy = arg.dy;
    }
    return *this;
}


bool
rectangle::empty(void)
    const
{
    return (dx <= 0 || dy <= 0);
}


rectangle
rectangle::operator*(const rectangle &rhs)
    const
{
    rectangle result(*this);
    result *= rhs;
    return result;
}


void
rectangle::clear(void)
{
    x = 0;
    y = 0;
    dx = 0;
    dy = 0;
}


static inline double
dmin(double a, double b)
{
    return (a < b ? a : b);
}


static inline double
dmax(double a, double b)
{
    return (a > b ? a : b);
}


rectangle &
rectangle::operator*=(const rectangle &rhs)
{
    if (empty())
        return *this;
    if (rhs.empty())
    {
        clear();
        return *this;
    }

    double lhs_x_lo = x;
    double lhs_x_hi = x + dx;
    double rhs_x_lo = rhs.x;
    double rhs_x_hi = rhs.x + rhs.dx;

    double x_lo = dmax(lhs_x_lo, rhs_x_lo);
    double x_hi = dmin(lhs_x_hi, rhs_x_hi);

    double lhs_y_lo = y;
    double lhs_y_hi = y + dy;
    double rhs_y_lo = rhs.y;
    double rhs_y_hi = rhs.y + rhs.dy;

    double y_lo = dmax(lhs_y_lo, rhs_y_lo);
    double y_hi = dmin(lhs_y_hi, rhs_y_hi);

    if (x_lo >= x_hi || y_lo >= y_hi)
    {
        x = 0;
        y = 0;
        dx = 0;
        dy = 0;
    }
    else
    {
        x = x_lo;
        y = y_lo;
        dx = x_hi - x_lo;
        dy = y_hi - y_lo;
    }

    return *this;
}


std::string
rectangle::representation(void)
    const
{
    return string_from_format("{ %.9g, %.9g, %.9g, %.9g }", x, y, dx, dy);
}


rectangle &
rectangle::operator+=(const rectangle &rhs)
{
    if (rhs.empty())
        return *this;
    if (empty())
    {
        *this = rhs;
        return *this;
    }
    double lhs_x_lo = x;
    double lhs_x_hi = x + dx;
    double rhs_x_lo = rhs.x;
    double rhs_x_hi = rhs.x + rhs.dx;

    double x_lo = dmin(lhs_x_lo, rhs_x_lo);
    double x_hi = dmax(lhs_x_hi, rhs_x_hi);

    double lhs_y_lo = y;
    double lhs_y_hi = y + dy;
    double rhs_y_lo = rhs.y;
    double rhs_y_hi = rhs.y + rhs.dy;

    double y_lo = dmin(lhs_y_lo, rhs_y_lo);
    double y_hi = dmax(lhs_y_hi, rhs_y_hi);

    if (x_lo >= x_hi || y_lo >= y_hi)
    {
        x = 0;
        y = 0;
        dx = 0;
        dy = 0;
    }
    else
    {
        x = x_lo;
        y = y_lo;
        dx = x_hi - x_lo;
        dy = y_hi - y_lo;
    }

    return *this;
}


rectangle
rectangle::operator+(const rectangle &rhs)
    const
{
    rectangle result = *this;
    result += rhs;
    return result;
}


bool
rectangle::operator==(const rectangle &rhs)
    const
{
    if (empty() && rhs.empty())
        return true;
    return (x == rhs.x && y == rhs.y && dx == rhs.dx && dy == rhs.dy);
}


int
rectangle::idx(void)
    const
{
    return int(ceil(dx));
}


int
rectangle::idy(void)
    const
{
    return int(ceil(dy));
}


bool
rectangle::contains(double px, double py)
    const
{
    return (x <= px && px < x + dx && y <= py && py < y + dy);
}


// vim: set ts=8 sw=4 et :
