//
// itractorgen - generate images of tractors
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_RECTANGLE_H
#define LIB_RECTANGLE_H

#include <lib/config.h>
#include <string>

/**
  * The rectangle class is used to represent a 2D area.
  */
class rectangle
{
public:
    double x, y, dx, dy;

    /**
      * The destructor.
      */
    virtual ~rectangle();

    /**
      * The default constructor.
      */
    rectangle();

    /**
      * The constructor.
      */
    rectangle(double x, double y, double dx, double dy);

    /**
      * The copy constructor.
      */
    rectangle(const rectangle &);

    /**
      * The assignment operator.
      */
    rectangle &operator=(const rectangle &);

    /**
      * The contains method may be used to determine whether or not this
      * rectangle contains the giben point.
      */
    bool contains(double px, double py) const;

    /**
      * The empty method is used to determine whether or not this
      * rectangle is empty (no width or no height).
      */
    bool empty(void) const;

    /**
      * The bool operator returns true of this rectangle isn't empty, or
      * false if it is empty.
      */
    operator bool (void) const { return !empty(); }

    /**
      * The not operator returns true of this rectangle is empty, or
      * false if it is not empty.
      */
    bool operator ! () const { return empty(); }

    /**
      * The in-situ multiply operator is used to intersect this
      * rectangle with another rectangle.  If there is no overlap, the
      * result is empty.
      *
      * @param rhs
      *     The other rectangle to intersect with.
      */
    rectangle &operator*=(const rectangle &rhs);

    /**
      * The multiply operator is used to intersect this rectangle with
      * another rectangle.  If there is no overlap, the result is
      * empty.
      *
      * @param rhs
      *     The other rectangle to intersect with.
      */
    rectangle operator*(const rectangle &rhs) const;

    /**
      * The in-situ addition operator is used to union this
      * rectangle with another rectangle.
      *
      * @param rhs
      *     The other rectangle to union with.
      */
    rectangle &operator+=(const rectangle &rhs);

    /**
      * The addition operator is used to union this rectangle with
      * another rectangle.
      *
      * @param rhs
      *     The other rectangle to union with.
      */
    rectangle operator+(const rectangle &rhs) const;

    /**
      * The clear method may be used to set this rectangle to be
      * completely empoty (no width, no height).
      */
    void clear(void);

    /**
      * The representation method is used to obtain a textual representation
      * of this rectangle, usually for debugging purposes.
      */
    std::string representation(void) const;

    /**
      * The equality operator is used to test two rectangles for
      * equality.
      *
      * @param rhs
      *     The rights hand side of the comparison (*this is the left
      *     hand side)
      */
    bool operator==(const rectangle &rhs) const;

    /**
      * The inequality operator is used to test two rectangles for
      * inequality.
      *
      * @param rhs
      *     The rights hand side of the comparison (*this is the left
      *     hand side)
      */
    bool operator!=(const rectangle &rhs) const { return !operator==(rhs); }

    /**
      * The idx method may be used to ovtain the value of int(ceil(dx)).
      */
    int idx(void) const;

    /**
      * The idy method may be used to ovtain the value of int(ceil(dy)).
      */
    int idy(void) const;
};

// vim: set ts=8 sw=4 et :
#endif // LIB_RECTANGLE_H
